package canalrun.com.crgeorot2;

import android.content.Context;
import android.hardware.GeomagneticField;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.location.Location;
import android.location.LocationListener;
import android.location.LocationManager;
import android.os.Bundle;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.BA.*;

@Version(1.16f)
@ShortName("CRGeoRot2")
@Events(values={"Bearing(bearing as Double)", "LogMsg(Msg as String)"})
@DontInheritEvents
// @DependsOn(values={""})
@ActivityObject

public class CRGeoRot2 implements SensorEventListener, LocationListener {
	private String eventName;
	private BA ba;
	
	public static final String FIXED = "FIXED"; // location min time
	private static final int LOCATION_MIN_TIME = 30 * 1000;
	private static final int LOCATION_MIN_DISTANCE = 10;
	
	private float[] mRotationMatrix = new float[16];
	private float[] mOrientation = new float[3];
	
	private SensorManager sensorManager;
	private Sensor sensorRotation;
	private LocationManager locationManager;
	
	private Location currentLocation;
	private GeomagneticField geomagneticField;
	private double mDeclination;
	
	private double bearing = 0;
	private double initialLat = 40.758896;
	private double initialLon = -73.985130;

	private boolean Inited = false;
	
	private class IIRClass {
		boolean enable = false;
		boolean firsttime = true;
		double Y = 0.0; // Y(n-1) old value
		double c0 = 0.90; // old val coeff
		double c1 = 0.10; // new val coeff
		double thresh = 0.20; // threshold for fast update
	  
		double calc(double X) { // pass X(n) new value
			if (firsttime == true) {
				Y = X;
				
				MyLogMsg("IIRClass first " + firsttime + ", enb " + enable  + ", X " + X + ", Y " + Y + ", c0 " + c0 + ", c1 " + c1);
				
				firsttime = false;
			}
		  
			if (enable == true) {
				if (Math.abs(X - Y) < (thresh * Y)) {
					Y = (Y * c0 + X * c1);
				} else {
					Y = X;
				}
			} else {
				Y = X;
			}
		  
//			MyLogMsg("IIRClass.calc " + ", X" + X + ", Y" + Y + ", firsttime" + firsttime);

			return(Y);
		}
	}
	
	private IIRClass IIRFilter = new IIRClass();
	
    private void MyLogMsg(String msg) {
		ba.raiseEventFromUI(this, eventName + "_logmsg", msg);
    }

    public String Identify() {
    	Float ver = this.getClass().getAnnotation(anywheresoftware.b4a.BA.Version.class).value();
    	
        return("CRGeoRot2 B4A Library ver: " + ver);
    }
    
    public boolean isInitialized() {
        return(Inited);
    }
    
    public void SetInitialLLocation(double lat, double lon) {
    	initialLat = lat;
    	initialLon = lon;
    }
    
    public void setIIR(boolean enable, double c0, double c1, double thresh) {
    	IIRFilter.enable = enable;
    	IIRFilter.c0 = c0;
    	IIRFilter.c1 = c1;
    	IIRFilter.thresh = thresh;    	
    }
    
    
	public void Initialize(final BA ba, String evname) {
		this.ba = ba;
		
		this.eventName = evname.toLowerCase(BA.cul);
		
		sensorManager = (SensorManager) ba.context.getSystemService(Context.SENSOR_SERVICE);
		
		sensorRotation = sensorManager.getDefaultSensor(Sensor.TYPE_ROTATION_VECTOR);

		// listen to these sensors
		sensorManager.registerListener(this, sensorRotation, SensorManager.SENSOR_DELAY_NORMAL);

		// location manager from system service
		locationManager = (LocationManager) ba.context.getSystemService(Context.LOCATION_SERVICE);
	 
		// request location data
		locationManager.requestLocationUpdates(LocationManager.GPS_PROVIDER, LOCATION_MIN_TIME, LOCATION_MIN_DISTANCE, this);
	 
		// get last known position
		Location gpsLocation = locationManager.getLastKnownLocation(LocationManager.GPS_PROVIDER);
	 
		if (gpsLocation != null) {
			currentLocation = gpsLocation;
		} else {
			// try with network provider
			Location networkLocation = locationManager.getLastKnownLocation(LocationManager.NETWORK_PROVIDER);
	 
			if (networkLocation != null) {
				currentLocation = networkLocation;
			} else {
				// Fix a position
				currentLocation = new Location(FIXED);
				currentLocation.setAltitude(1);
				currentLocation.setLatitude(initialLat);
				currentLocation.setLongitude(initialLon);
			}
		}

		// set current location, init mDeclination
		onLocationChanged(currentLocation);

		Inited = true;
		
        MyLogMsg("CRGeoRot2 Initialize - EventName:" + evname + ", Lat:" +
        		currentLocation.getLatitude() + ", Lon:" + currentLocation.getLongitude() + ", Package:" + BA.packageName);
	}

	public void Destroy() {
		// remove listeners
		sensorManager.unregisterListener(this, sensorRotation);
		locationManager.removeUpdates(this);
		
        MyLogMsg("CRGeoRot2 Destroy");
	}
	 
	@Override
	public void onLocationChanged(Location location) {
		currentLocation = location;
	    
		geomagneticField = new GeomagneticField(
				(float) currentLocation.getLatitude(),
				(float) currentLocation.getLongitude(),
				(float) currentLocation.getAltitude(),
				System.currentTimeMillis());

		// getDeclination returns degrees
		mDeclination = geomagneticField.getDeclination();

//		MyLogMsg("CRGeoRot2 onLocationChanged, mDeclination: " + mDeclination);
	}
	 
	@Override
	public void onStatusChanged(String provider, int status, Bundle extras) {
	}
	 
	@Override
	public void onProviderEnabled(String provider) {
	}
	 
	@Override
	public void onProviderDisabled(String provider) {
	}
	 
	@Override
	public void onSensorChanged(SensorEvent event) {
		if(event.sensor.getType() == Sensor.TYPE_ROTATION_VECTOR) {
			SensorManager.getRotationMatrixFromVector(mRotationMatrix , event.values);
			SensorManager.getOrientation(mRotationMatrix, mOrientation);
			
			bearing = Math.toDegrees(mOrientation[0]) + mDeclination;
			
			if (bearing < 0) bearing += 360; // bearing should be in range 0-360
			
			bearing = IIRFilter.calc(bearing);
			
			ba.raiseEventFromUI(this, eventName + "_bearing", bearing);
		}

//		MyLogMsg("CRGeoRot2 onSensorChanged");
	}
    
	@Override
	public void onAccuracyChanged(Sensor sensor, int accuracy) {
		if (sensor.getType() == Sensor.TYPE_MAGNETIC_FIELD && accuracy == SensorManager.SENSOR_STATUS_UNRELIABLE) {
			// manage fact that compass data are unreliable ...
		}

//		MyLogMsg("CRGeoRot2 onAccuracyChanged");
	}
}
